@@ -24,6 +24,16 @@
 #define FUSE_USE_VERSION 35
 #include <fuse_lowlevel.h>
 
+#ifdef HAVE_FREERDP3
+#define COMMON(x) common.x
+#else
+#define COMMON(x) x
+#endif
+
+#define FRDP_CLIPBOARD_FORMAT_PNG          0xD011
+#define FRDP_CLIPBOARD_FORMAT_JPEG         0xD012
+#define FRDP_CLIPBOARD_FORMAT_TEXT_URILIST 0xD014
+
 typedef struct
 {
   guchar   *data;
@@ -58,6 +68,13 @@ typedef struct
   FILEDESCRIPTORW *descriptor;
 } FrdpLocalFileInfo;
 
+typedef struct
+{
+  guint              clip_data_id;
+  gsize              local_files_count;
+  FrdpLocalFileInfo *local_files_infos;
+} FrdpLocalLockData;
+
 typedef struct _FrdpRemoteFileInfo FrdpRemoteFileInfo;
 
 struct _FrdpRemoteFileInfo
@@ -108,6 +125,14 @@ typedef struct
   GMutex                       fuse_mutex;
 
   fuse_ino_t                   current_inode;
+
+  GList                       *locked_data;           /* List of locked arrays of files - list of (FrdpLocalLockData *) */
+  GMutex                       lock_mutex;
+  gboolean                     pending_lock;          /* Lock was requested right after format list has been sent */
+  guint                        pending_lock_id;       /* Id for the pending lock */
+  gboolean                     awaiting_data_request; /* Format list has been send but data were not requested yet */
+
+  guint                        remote_clip_data_id;   /* clipDataId for copying from remote side */
 } FrdpChannelClipboardPrivate;
 
 G_DEFINE_TYPE_WITH_PRIVATE (FrdpChannelClipboard, frdp_channel_clipboard, FRDP_TYPE_CHANNEL)
@@ -129,6 +154,10 @@ static void  clipboard_owner_change_cb                 (GtkClipboard         *cl
                                                         GdkEventOwnerChange  *event,
                                                         gpointer              user_data);
 
+static void  frdp_local_lock_data_free                 (FrdpLocalLockData    *lock_data);
+static void  lock_current_local_files                  (FrdpChannelClipboard *self,
+                                                        guint                 clip_data_id);
+
 static void
 frdp_channel_clipboard_get_property (GObject    *object,
                                      guint       property_id,
@@ -181,7 +210,15 @@ frdp_channel_clipboard_finalize (GObject *object)
   _gtk_clipboard_clear_func (priv->gtk_clipboard, self);
   g_clear_pointer (&priv->fuse_directory, g_free);
 
+  g_mutex_lock (&priv->lock_mutex);
+
+  g_list_free_full (priv->locked_data, (GDestroyNotify) frdp_local_lock_data_free);
+  priv->locked_data = NULL;
+
+  g_mutex_unlock (&priv->lock_mutex);
+
   g_mutex_clear (&priv->fuse_mutex);
+  g_mutex_clear (&priv->lock_mutex);
 
   G_OBJECT_CLASS (frdp_channel_clipboard_parent_class)->finalize (object);
 }
@@ -269,6 +306,8 @@ request_size (FrdpChannelClipboard *self,
   file_contents_request.cbRequested = 8;
   file_contents_request.nPositionHigh = 0;
   file_contents_request.nPositionLow = 0;
+  file_contents_request.haveClipDataId = TRUE;
+  file_contents_request.clipDataId = priv->remote_clip_data_id;
 
   size_request = g_new0 (FrdpRemoteFileRequest, 1);
   size_request->index = index;
@@ -435,7 +474,8 @@ fuse_read (fuse_req_t             request,
       file_contents_request.cbRequested = size;
       file_contents_request.nPositionHigh = offset >> 32;
       file_contents_request.nPositionLow = offset & 0xffffffff;
-      file_contents_request.haveClipDataId = FALSE;
+      file_contents_request.haveClipDataId = TRUE;
+      file_contents_request.clipDataId = priv->remote_clip_data_id;
 
       data_request = g_new0 (FrdpRemoteFileRequest, 1);
       data_request->index = index;
@@ -610,8 +650,11 @@ frdp_channel_clipboard_init (FrdpChannelClipboard *self)
 
   priv->gtk_clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
   priv->clipboard_owner_changed_id = g_signal_connect (priv->gtk_clipboard, "owner-change", G_CALLBACK (clipboard_owner_change_cb), self);
-  priv->fgdw_id = CB_FORMAT_TEXTURILIST;
+  priv->fgdw_id = FRDP_CLIPBOARD_FORMAT_TEXT_URILIST;
   priv->current_inode = FUSE_ROOT_ID + 1;
+  priv->locked_data = NULL;
+  priv->pending_lock = FALSE;
+  priv->remote_clip_data_id = 0;
 
   argv[0] = "gnome-connections";
   argv[1] = "-d";
@@ -621,6 +664,7 @@ frdp_channel_clipboard_init (FrdpChannelClipboard *self)
   priv->remote_files_requests = g_hash_table_new (g_direct_hash, g_direct_equal);
 
   g_mutex_init (&priv->fuse_mutex);
+  g_mutex_init (&priv->lock_mutex);
 
   priv->fuse_directory = g_mkdtemp (g_strdup_printf ("%s/clipboard-XXXXXX/", g_get_user_runtime_dir ()));
 
@@ -669,6 +713,7 @@ send_client_capabilities (FrdpChannelClipboard *self)
   general_capability_set.version = CB_CAPS_VERSION_2;
   general_capability_set.generalFlags = CB_USE_LONG_FORMAT_NAMES |
                                         CB_STREAM_FILECLIP_ENABLED |
+                                        CB_CAN_LOCK_CLIPDATA |
                                         CB_FILECLIP_NO_FILE_PATHS |
                                         CB_HUGE_FILE_SUPPORT_ENABLED;
 
@@ -705,10 +750,10 @@ send_client_format_list (FrdpChannelClipboard *self)
         formats[j].formatId = CF_TEXT;
         formats[j++].formatName = NULL;
       } else if (g_strcmp0 (atom_name, "image/png") == 0) {
-        formats[j].formatId = CB_FORMAT_PNG;
+        formats[j].formatId = FRDP_CLIPBOARD_FORMAT_PNG;
         formats[j++].formatName = NULL;
       } else if (g_strcmp0 (atom_name, "image/jpeg") == 0) {
-        formats[j].formatId = CB_FORMAT_JPEG;
+        formats[j].formatId = FRDP_CLIPBOARD_FORMAT_JPEG;
         formats[j++].formatName = NULL;
       } else if (g_strcmp0 (atom_name, "image/bmp") == 0) {
         formats[j].formatId = CF_DIB;
@@ -722,11 +767,12 @@ send_client_format_list (FrdpChannelClipboard *self)
     }
   }
 
-  format_list.msgType = CB_FORMAT_LIST;
-  format_list.msgFlags = CB_RESPONSE_OK;
+  format_list.COMMON(msgType) = CB_FORMAT_LIST;
+  format_list.COMMON(msgFlags) = CB_RESPONSE_OK;
   format_list.numFormats = j;
   format_list.formats = formats;
 
+  priv->awaiting_data_request = TRUE;
   ret = priv->cliprdr_client_context->ClientFormatList (priv->cliprdr_client_context, &format_list);
 
   if (formats != NULL) {
@@ -859,9 +905,9 @@ send_client_format_list_response (FrdpChannelClipboard *self,
   CLIPRDR_FORMAT_LIST_RESPONSE  response = { 0 };
   FrdpChannelClipboardPrivate  *priv = frdp_channel_clipboard_get_instance_private (self);
 
-  response.msgType = CB_FORMAT_LIST_RESPONSE;
-  response.msgFlags = status ? CB_RESPONSE_OK : CB_RESPONSE_FAIL;
-  response.dataLen = 0;
+  response.COMMON(msgType) = CB_FORMAT_LIST_RESPONSE;
+  response.COMMON(msgFlags) = status ? CB_RESPONSE_OK : CB_RESPONSE_FAIL;
+  response.COMMON(dataLen) = 0;
 
   return priv->cliprdr_client_context->ClientFormatListResponse (priv->cliprdr_client_context, &response);
 }
@@ -879,6 +925,39 @@ replace_ascii_character (gchar *text,
   }
 }
 
+static WCHAR *
+convert_to_unicode (const gchar *text)
+{
+  WCHAR *result = NULL;
+
+  if (text != NULL) {
+#ifdef HAVE_FREERDP3
+    result = ConvertUtf8ToWCharAlloc (text, NULL);
+#else
+    ConvertToUnicode (CP_UTF8, 0, (LPCSTR) text, -1, &result, 0);
+#endif
+  }
+
+  return result;
+}
+
+static gchar *
+convert_from_unicode (const WCHAR *text,
+                      gint         text_length)
+{
+  gchar *result = NULL;
+
+  if (text != NULL) {
+#ifdef HAVE_FREERDP3
+    result = ConvertWCharNToUtf8Alloc (text, text_length, NULL);
+#else
+    ConvertFromUnicode (CP_UTF8, 0, text, text_length, &result, 0, NULL, NULL);
+#endif
+  }
+
+  return result;
+}
+
 /* TODO: Rewrite this using async methods of GtkCLipboard once we move to Gtk4 */
 static void
 _gtk_clipboard_get_func (GtkClipboard     *clipboard,
@@ -886,12 +965,19 @@ _gtk_clipboard_get_func (GtkClipboard     *clipboard,
                          guint             info,
                          gpointer          user_data)
 {
+  CLIPRDR_LOCK_CLIPBOARD_DATA  lock_clipboard_data = { 0 };
   FrdpChannelClipboard        *self = (FrdpChannelClipboard *) user_data;
   FrdpChannelClipboardPrivate *priv = frdp_channel_clipboard_get_instance_private (self);
   FrdpClipboardRequest        *current_request;
   gchar                       *data = NULL;
   gint                         length;
 
+  lock_clipboard_data.COMMON(msgType) = CB_LOCK_CLIPDATA;
+  lock_clipboard_data.COMMON(msgFlags) = 0;
+  lock_clipboard_data.COMMON(dataLen) = 4;
+  lock_clipboard_data.clipDataId = ++priv->remote_clip_data_id;
+  priv->cliprdr_client_context->ClientLockClipboardData (priv->cliprdr_client_context, &lock_clipboard_data);
+
   current_request = frdp_clipboard_request_send (self, info);
   if (current_request != NULL) {
 
@@ -900,13 +986,15 @@ _gtk_clipboard_get_func (GtkClipboard     *clipboard,
 
     if (info == CF_UNICODETEXT) {
       /* TODO - convert CR LF to CR */
-      length = ConvertFromUnicode (CP_UTF8, 0, (WCHAR *) current_request->responses[0].data, (int) (current_request->responses[0].length / sizeof (WCHAR)), &data, 0, NULL, NULL);
-
-      gtk_selection_data_set (selection_data,
-                              gdk_atom_intern ("UTF8_STRING", FALSE),
-                              8,
-                              (guchar *) data,
-                              length);
+      data = convert_from_unicode ((WCHAR *) current_request->responses[0].data, current_request->responses[0].length / sizeof (WCHAR));
+      if (data != NULL) {
+        length = strlen (data);
+        gtk_selection_data_set (selection_data,
+                                gdk_atom_intern ("UTF8_STRING", FALSE),
+                                8,
+                                (guchar *) data,
+                                length);
+      }
     } else if (info == CF_DIB) {
       /* This has been inspired by function transmute_cf_dib_to_image_bmp() from gtk */
       BITMAPINFOHEADER *bi = (BITMAPINFOHEADER *) current_request->responses[0].data;
@@ -945,7 +1033,7 @@ _gtk_clipboard_get_func (GtkClipboard     *clipboard,
           priv->remote_files_infos = g_new0 (FrdpRemoteFileInfo, priv->remote_files_count);
 
           for (i = 0; i < count; i++) {
-            length = ConvertFromUnicode (CP_UTF8, 0, (WCHAR *) files[i].cFileName, (int) (260 / sizeof (WCHAR)), &path, 0, NULL, NULL);
+            path = convert_from_unicode ((WCHAR *) files[i].cFileName, 260 / sizeof (WCHAR));
 
             replace_ascii_character (path, '\\', '/');
 
@@ -1051,9 +1139,10 @@ static void
 _gtk_clipboard_clear_func (GtkClipboard *clipboard,
                            gpointer      user_data)
 {
-  FrdpChannelClipboard        *self = (FrdpChannelClipboard *) user_data;
-  FrdpChannelClipboardPrivate *priv = frdp_channel_clipboard_get_instance_private (self);
-  guint                        i;
+  CLIPRDR_UNLOCK_CLIPBOARD_DATA  unlock_clipboard_data = { 0 };
+  FrdpChannelClipboard          *self = (FrdpChannelClipboard *) user_data;
+  FrdpChannelClipboardPrivate   *priv = frdp_channel_clipboard_get_instance_private (self);
+  guint                          i;
 
   g_mutex_lock (&priv->fuse_mutex);
 
@@ -1070,6 +1159,12 @@ _gtk_clipboard_clear_func (GtkClipboard *clipboard,
 
   g_mutex_unlock (&priv->fuse_mutex);
 
+  unlock_clipboard_data.COMMON(msgType) = CB_UNLOCK_CLIPDATA;
+  unlock_clipboard_data.COMMON(msgFlags) = 0;
+  unlock_clipboard_data.COMMON(dataLen) = 4;
+  unlock_clipboard_data.clipDataId = priv->remote_clip_data_id;
+  priv->cliprdr_client_context->ClientUnlockClipboardData (priv->cliprdr_client_context, &unlock_clipboard_data);
+
   clear_local_files_infos (self);
 
   priv->remote_data_in_clipboard = FALSE;
@@ -1113,7 +1208,7 @@ server_format_list (CliprdrClientContext      *context,
           atom = gdk_atom_intern ("UTF8_STRING", FALSE);
         } else if (format_list->formats[i].formatId == CF_DIB) {
           atom = gdk_atom_intern ("image/bmp", FALSE);
-        } else if (format_list->formats[i].formatId == CB_FORMAT_PNG) {
+        } else if (format_list->formats[i].formatId == FRDP_CLIPBOARD_FORMAT_PNG) {
           atom = gdk_atom_intern ("image/png", FALSE);
         }
 
@@ -1186,9 +1281,11 @@ frdp_local_file_info_new (GFile     *file,
     relative_path = g_file_get_relative_path (root, file);
     replace_ascii_character (relative_path, '/', '\\');
 
-    ConvertToUnicode (CP_UTF8, 0, (LPCSTR) relative_path, -1, &file_name, 0);
-    memcpy (frdp_file_info->descriptor->cFileName, file_name, strlen (relative_path) * 2);
-    g_free (file_name);
+    file_name = convert_to_unicode (relative_path);
+    if (file_name != NULL) {
+      memcpy (frdp_file_info->descriptor->cFileName, file_name, strlen (relative_path) * 2);
+      g_free (file_name);
+    }
     g_free (relative_path);
 
     file_size = g_file_info_get_size (file_info);
@@ -1260,8 +1357,8 @@ send_data_response (FrdpChannelClipboard *self,
   if (size > UINT32_MAX)
     return ERROR_INVALID_PARAMETER;
 
-  response.msgFlags = (data) ? CB_RESPONSE_OK : CB_RESPONSE_FAIL;
-  response.dataLen = (guint32) size;
+  response.COMMON(msgFlags) = (data) ? CB_RESPONSE_OK : CB_RESPONSE_FAIL;
+  response.COMMON(dataLen) = (guint32) size;
   response.requestedFormatData = data;
 
   return priv->cliprdr_client_context->ClientFormatDataResponse (priv->cliprdr_client_context, &response);
@@ -1285,14 +1382,17 @@ clipboard_content_received (GtkClipboard     *clipboard,
   length = gtk_selection_data_get_length (selection_data);
   data_type = gtk_selection_data_get_data_type (selection_data);
 
-  if (length >= 0) {
+  if (length > 0) {
     if (data_type == gdk_atom_intern ("UTF8_STRING", FALSE)) {
       text = gtk_selection_data_get_text (selection_data);
       text_length = strlen ((gchar *) text);
-      if (ConvertToUnicode (CP_UTF8, 0, (LPCSTR) text, text_length, (WCHAR **) &data, 0) > 0) {
+
+      data = (guchar *) convert_to_unicode ((gchar *) text);
+      if (data != NULL) {
         send_data_response (self, data, (text_length + 1) * sizeof (WCHAR));
         g_free (data);
       }
+
       g_free (text);
     } else if (data_type == gdk_atom_intern ("image/png", FALSE)) {
       pixbuf = gtk_selection_data_get_pixbuf (selection_data);
@@ -1373,6 +1473,12 @@ clipboard_content_received (GtkClipboard     *clipboard,
       }
       g_list_free_full (list, g_free);
 
+      if (priv->awaiting_data_request && priv->pending_lock) {
+        lock_current_local_files (self, priv->pending_lock_id);
+
+        priv->awaiting_data_request = FALSE;
+      }
+
       send_data_response (self, data, priv->local_files_count * sizeof (FILEDESCRIPTORW) + 4);
     }
   } else {
@@ -1398,13 +1504,13 @@ server_format_data_request (CliprdrClientContext              *context,
                                       clipboard_content_received,
                                       self);
       break;
-    case CB_FORMAT_PNG:
+    case FRDP_CLIPBOARD_FORMAT_PNG:
       gtk_clipboard_request_contents (priv->gtk_clipboard,
                                       gdk_atom_intern ("image/png", FALSE),
                                       clipboard_content_received,
                                       self);
       break;
-    case CB_FORMAT_JPEG:
+    case FRDP_CLIPBOARD_FORMAT_JPEG:
       gtk_clipboard_request_contents (priv->gtk_clipboard,
                                       gdk_atom_intern ("image/jpeg", FALSE),
                                       clipboard_content_received,
@@ -1445,7 +1551,7 @@ server_format_data_response (CliprdrClientContext               *context,
     self = (FrdpChannelClipboard *) context->custom;
     priv = frdp_channel_clipboard_get_instance_private (self);
 
-    if (response->msgType == CB_FORMAT_DATA_RESPONSE) {
+    if (response->COMMON(msgType) == CB_FORMAT_DATA_RESPONSE) {
       if (priv->requests != NULL) {
         current_request = priv->requests->data;
         for (j = 0; j < current_request->count; j++) {
@@ -1457,10 +1563,10 @@ server_format_data_response (CliprdrClientContext               *context,
 
         if (subrequest_index >= 0 && subrequest_index < current_request->count) {
           current_request->responses[subrequest_index].handled = TRUE;
-          if (response->msgFlags & CB_RESPONSE_OK) {
-            current_request->responses[subrequest_index].length = response->dataLen;
-            current_request->responses[subrequest_index].data = g_new (guchar, response->dataLen);
-            memcpy (current_request->responses[subrequest_index].data, response->requestedFormatData, response->dataLen);
+          if (response->COMMON(msgFlags) & CB_RESPONSE_OK) {
+            current_request->responses[subrequest_index].length = response->COMMON(dataLen);
+            current_request->responses[subrequest_index].data = g_new (guchar, response->COMMON(dataLen));
+            memcpy (current_request->responses[subrequest_index].data, response->requestedFormatData, response->COMMON(dataLen));
           } else {
             g_warning ("Clipboard data request failed!");
           }
@@ -1481,22 +1587,50 @@ server_file_contents_request (CliprdrClientContext                *context,
   FrdpChannelClipboard           *self = (FrdpChannelClipboard *) context->custom;
   FrdpChannelClipboardPrivate    *priv = frdp_channel_clipboard_get_instance_private (self);
   CLIPRDR_FILE_CONTENTS_RESPONSE  response = { 0 };
+  FrdpLocalFileInfo               local_file_info;
+  FrdpLocalLockData              *ldata;
   GFileInputStream               *stream;
   GFileInfo                      *file_info;
   GFileType                       file_type;
+  gboolean                        local_file_info_set = FALSE, clip_data_id_found = FALSE;
   guint64                        *size;
   goffset                         offset;
   guchar                         *data = NULL;
   gssize                          bytes_read;
+  GList                          *iter;
   GFile                          *file;
 
-  response.msgType = CB_FILECONTENTS_RESPONSE;
-  response.msgFlags = CB_RESPONSE_FAIL;
+  response.COMMON(msgType) = CB_FILECONTENTS_RESPONSE;
+  response.COMMON(msgFlags) = CB_RESPONSE_FAIL;
   response.streamId = file_contents_request->streamId;
 
+  g_mutex_lock (&priv->lock_mutex);
+
+  if (file_contents_request->haveClipDataId) {
+    for (iter = priv->locked_data; iter != NULL; iter = iter->next) {
+      ldata = (FrdpLocalLockData *) iter->data;
+
+      if (ldata->clip_data_id == file_contents_request->clipDataId) {
+        clip_data_id_found = TRUE;
+        if (file_contents_request->listIndex < ldata->local_files_count) {
+          local_file_info = ldata->local_files_infos[file_contents_request->listIndex];
+          local_file_info_set = TRUE;
+        }
+        break;
+      }
+    }
+  }
+
+  if (!local_file_info_set && !clip_data_id_found) {
+    if (file_contents_request->listIndex < priv->local_files_count) {
+      local_file_info = priv->local_files_infos[file_contents_request->listIndex];
+      local_file_info_set = TRUE;
+    }
+  }
+
   /* TODO: Make it async. Signal progress if FD_SHOWPROGRESSUI is present. */
-  if (file_contents_request->listIndex < priv->local_files_count) {
-    file = g_file_new_for_uri (priv->local_files_infos[file_contents_request->listIndex].uri);
+  if (local_file_info_set) {
+    file = g_file_new_for_uri (local_file_info.uri);
 
     if (file_contents_request->dwFlags & FILECONTENTS_SIZE) {
       file_info = g_file_query_info (file, G_FILE_ATTRIBUTE_STANDARD_SIZE, G_FILE_QUERY_INFO_NONE, NULL, NULL);
@@ -1505,8 +1639,8 @@ server_file_contents_request (CliprdrClientContext                *context,
 
       response.requestedData = (guchar *) size;
       response.cbRequested = 8;
-      response.dataLen = 8;
-      response.msgFlags = CB_RESPONSE_OK;
+      response.COMMON(dataLen) = 8;
+      response.COMMON(msgFlags) = CB_RESPONSE_OK;
 
       g_object_unref (file_info);
     } else if (file_contents_request->dwFlags & FILECONTENTS_RANGE) {
@@ -1524,8 +1658,8 @@ server_file_contents_request (CliprdrClientContext                *context,
 
           response.requestedData = data;
           response.cbRequested = bytes_read;
-          response.dataLen = bytes_read;
-          response.msgFlags = CB_RESPONSE_OK;
+          response.COMMON(dataLen) = bytes_read;
+          response.COMMON(msgFlags) = CB_RESPONSE_OK;
         }
       } else {
         g_warning ("Content of a directory was requested!");
@@ -1540,6 +1674,8 @@ server_file_contents_request (CliprdrClientContext                *context,
     g_warning ("Requested index is outside of the file list!");
   }
 
+  g_mutex_unlock (&priv->lock_mutex);
+
   return priv->cliprdr_client_context->ClientFileContentsResponse (priv->cliprdr_client_context, &response);
 }
 
@@ -1553,7 +1689,7 @@ server_file_contents_response (CliprdrClientContext                 *context,
   FrdpRemoteFileRequest       *request;
   struct stat                  attr = {0};
 
-  if (context != NULL && file_contents_response->msgFlags & CB_RESPONSE_OK) {
+  if (context != NULL && file_contents_response->COMMON(msgFlags) & CB_RESPONSE_OK) {
     self = (FrdpChannelClipboard *) context->custom;
     priv = frdp_channel_clipboard_get_instance_private (self);
 
@@ -1597,11 +1733,99 @@ server_file_contents_response (CliprdrClientContext                 *context,
       g_free (request);
       g_mutex_unlock (&priv->fuse_mutex);
     }
+  } else {
+    if (file_contents_response->COMMON(msgFlags) & CB_RESPONSE_FAIL) {
+      g_warning ("Server file response has failed!");
+    }
   }
 
   return CHANNEL_RC_OK;
 }
 
+static void
+lock_current_local_files (FrdpChannelClipboard *self,
+                          guint                 clip_data_id)
+{
+  FrdpChannelClipboardPrivate *priv = frdp_channel_clipboard_get_instance_private (self);
+  FrdpLocalLockData           *lock_data;
+  guint                        i;
+
+  g_mutex_lock (&priv->lock_mutex);
+
+  /* TODO: Implement flock */
+  if (priv->local_files_count > 0) {
+    lock_data = g_new (FrdpLocalLockData, 1);
+    lock_data->clip_data_id = clip_data_id;
+    lock_data->local_files_count = priv->local_files_count;
+    lock_data->local_files_infos = g_new (FrdpLocalFileInfo, lock_data->local_files_count);
+    for (i = 0; i < lock_data->local_files_count; i++) {
+      lock_data->local_files_infos[i].descriptor = priv->local_files_infos[i].descriptor;
+      lock_data->local_files_infos[i].uri = g_strdup (priv->local_files_infos[i].uri);
+    }
+
+    priv->locked_data = g_list_append (priv->locked_data, lock_data);
+    if (priv->pending_lock_id == clip_data_id)
+      priv->pending_lock = FALSE;
+  }
+
+  g_mutex_unlock (&priv->lock_mutex);
+}
+
+static guint
+server_lock_clipboard_data (CliprdrClientContext              *context,
+                            const CLIPRDR_LOCK_CLIPBOARD_DATA *lock_clipboard_data)
+{
+  FrdpChannelClipboard        *self = (FrdpChannelClipboard *) context->custom;
+  FrdpChannelClipboardPrivate *priv = frdp_channel_clipboard_get_instance_private (self);
+
+  if (priv->awaiting_data_request) {
+    priv->pending_lock = TRUE;
+    priv->pending_lock_id = lock_clipboard_data->clipDataId;
+  } else {
+    lock_current_local_files (self, lock_clipboard_data->clipDataId);
+  }
+
+  return CHANNEL_RC_OK;
+}
+
+static void
+frdp_local_lock_data_free (FrdpLocalLockData *lock_data)
+{
+  guint i;
+
+  for (i = 0; i < lock_data->local_files_count; i++)
+    g_free (lock_data->local_files_infos[i].uri);
+  g_free (lock_data->local_files_infos);
+  g_free (lock_data);
+}
+
+static guint
+server_unlock_clipboard_data (CliprdrClientContext                *context,
+                              const CLIPRDR_UNLOCK_CLIPBOARD_DATA *unlock_clipboard_data)
+{
+  FrdpChannelClipboard        *self = (FrdpChannelClipboard *) context->custom;
+  FrdpChannelClipboardPrivate *priv = frdp_channel_clipboard_get_instance_private (self);
+  FrdpLocalLockData           *lock_data;
+  GList                       *iter;
+
+  g_mutex_lock (&priv->lock_mutex);
+
+  for (iter = priv->locked_data; iter != NULL; iter = iter->next) {
+    lock_data = iter->data;
+
+    if (lock_data->clip_data_id == unlock_clipboard_data->clipDataId) {
+      frdp_local_lock_data_free (lock_data);
+
+      priv->locked_data = g_list_delete_link (priv->locked_data, iter);
+      break;
+    }
+  }
+
+  g_mutex_unlock (&priv->lock_mutex);
+
+  return CHANNEL_RC_OK;
+}
+
 static void
 frdp_channel_clipboard_set_client_context (FrdpChannelClipboard *self,
                                            CliprdrClientContext *context)
@@ -1620,8 +1844,7 @@ frdp_channel_clipboard_set_client_context (FrdpChannelClipboard *self,
   context->ServerFileContentsRequest = server_file_contents_request;
   context->ServerFileContentsResponse = server_file_contents_response;
 
-  /* TODO: Implement these:
-       pcCliprdrServerLockClipboardData ServerLockClipboardData;
-       pcCliprdrServerUnlockClipboardData ServerUnlockClipboardData;
-   */
+  /* These don't lock/unlock files currently but store lists of files with their clipDataId. */
+  context->ServerLockClipboardData = server_lock_clipboard_data;
+  context->ServerUnlockClipboardData = server_unlock_clipboard_data;
 }
