@@ -28,6 +28,13 @@ struct _FrdpDisplayPrivate
 
   gboolean     allow_resize;
   gboolean     resize_supported;
+
+  gboolean     awaiting_certificate_verification;
+  gboolean     awaiting_certificate_change_verification;
+  gboolean     awaiting_authentication;
+
+  guint        certificate_verification_value;
+  guint        certificate_change_verification_value;
 };
 
 G_DEFINE_TYPE_WITH_PRIVATE (FrdpDisplay, frdp_display, GTK_TYPE_DRAWING_AREA)
@@ -39,7 +46,8 @@ enum
   PROP_PASSWORD,
   PROP_SCALING,
   PROP_ALLOW_RESIZE,
-  PROP_RESIZE_SUPPORTED
+  PROP_RESIZE_SUPPORTED,
+  PROP_DOMAIN
 };
 
 enum
@@ -49,6 +57,8 @@ enum
   RDP_DISCONNECTED,
   RDP_NEEDS_AUTHENTICATION,
   RDP_AUTH_FAILURE,
+  RDP_NEEDS_CERTIFICATE_VERIFICATION,
+  RDP_NEEDS_CERTIFICATE_CHANGE_VERIFICATION,
   LAST_SIGNAL
 };
 
@@ -171,23 +181,12 @@ frdp_display_scroll_event (GtkWidget      *widget,
       flags = FRDP_MOUSE_EVENT_HWHEEL;
       break;
     case GDK_SCROLL_SMOOTH:
-    /* Calculate delta and decide which event we have
-     * a delta X means horizontal, a delta Y means vertical scroll.
-     * Fixes https://bugzilla.gnome.org/show_bug.cgi?id=675959
-     */
-    if (event->delta_x > 0.5)
-      flags = FRDP_MOUSE_EVENT_HWHEEL;
-    else if (event->delta_x < -0.5)
-      flags = FRDP_MOUSE_EVENT_HWHEEL | FRDP_MOUSE_EVENT_WHEEL_NEGATIVE;
-    else if (event->delta_y > 0.5)
-      flags = FRDP_MOUSE_EVENT_WHEEL;
-    else if (event->delta_y < -0.5)
-      flags = FRDP_MOUSE_EVENT_WHEEL | FRDP_MOUSE_EVENT_WHEEL_NEGATIVE;
-    else {
-      g_debug ("scroll smooth unhandled");
-      return FALSE;
-    }
-    break;
+      frdp_session_mouse_smooth_scroll_event (priv->session,
+                                              event->x,
+                                              event->y,
+                                              event->delta_x,
+                                              event->delta_y);
+      return TRUE;
     default:
       return FALSE;
   }
@@ -293,6 +292,10 @@ frdp_display_get_property (GObject      *object,
         g_object_get (session, "password", &str_property, NULL);
         g_value_set_string (value, str_property);
         break;
+      case PROP_DOMAIN:
+        g_object_get (session, "domain", &str_property, NULL);
+        g_value_set_string (value, str_property);
+        break;
       case PROP_SCALING:
         g_object_get (session, "scaling", &str_property, NULL);
         g_value_set_boolean (value, (gboolean)GPOINTER_TO_INT (str_property));
@@ -327,6 +330,9 @@ frdp_display_set_property (GObject      *object,
       case PROP_PASSWORD:
         g_object_set (session, "password", g_value_get_string (value), NULL);
         break;
+      case PROP_DOMAIN:
+        g_object_set (session, "domain", g_value_get_string (value), NULL);
+        break;
       case PROP_SCALING:
         frdp_display_set_scaling (self, g_value_get_boolean (value));
         break;
@@ -377,6 +383,14 @@ frdp_display_class_init (FrdpDisplayClass *klass)
                                                         NULL,
                                                         G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE));
 
+  g_object_class_install_property (gobject_class,
+                                   PROP_DOMAIN,
+                                   g_param_spec_string ("domain",
+                                                        "domain",
+                                                        "domain",
+                                                        NULL,
+                                                        G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE));
+
   g_object_class_install_property (gobject_class,
                                    PROP_SCALING,
                                    g_param_spec_boolean ("scaling",
@@ -432,6 +446,35 @@ frdp_display_class_init (FrdpDisplayClass *klass)
                                             0, NULL, NULL, NULL,
                                             G_TYPE_NONE, 1,
                                             G_TYPE_STRING);
+
+  signals[RDP_NEEDS_CERTIFICATE_VERIFICATION] = g_signal_new ("rdp-needs-certificate-verification",
+                                                              G_TYPE_FROM_CLASS (klass),
+                                                              G_SIGNAL_RUN_LAST,
+                                                              0, NULL, NULL, NULL,
+                                                              G_TYPE_NONE, 7,
+                                                              G_TYPE_STRING,
+                                                              G_TYPE_UINT,
+                                                              G_TYPE_STRING,
+                                                              G_TYPE_STRING,
+                                                              G_TYPE_STRING,
+                                                              G_TYPE_STRING,
+                                                              G_TYPE_UINT);
+
+  signals[RDP_NEEDS_CERTIFICATE_CHANGE_VERIFICATION] = g_signal_new ("rdp-needs-certificate-change-verification",
+                                                                     G_TYPE_FROM_CLASS (klass),
+                                                                     G_SIGNAL_RUN_LAST,
+                                                                     0, NULL, NULL, NULL,
+                                                                     G_TYPE_NONE, 10,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_UINT,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_STRING,
+                                                                     G_TYPE_UINT);
 }
 
 static void
@@ -589,16 +632,188 @@ GtkWidget *frdp_display_new (void)
 }
 
 gboolean
-frdp_display_authenticate (FrdpDisplay *self,
-                           gchar **username,
-                           gchar **password,
-                           gchar **domain)
+frdp_display_authenticate (FrdpDisplay  *self,
+                           gchar       **username,
+                           gchar       **password,
+                           gchar       **domain)
 {
-  FrdpDisplayClass *klass = FRDP_DISPLAY_GET_CLASS (self);
+  FrdpDisplayPrivate *priv = frdp_display_get_instance_private (self);
+  GMainContext       *context;
 
   g_signal_emit (self, signals[RDP_NEEDS_AUTHENTICATION], 0);
 
-  return klass->authenticate (self, username, password, domain);
+  priv->awaiting_authentication = TRUE;
+
+  context = g_main_context_default ();
+
+  while (priv->awaiting_authentication)
+    g_main_context_iteration (context, FALSE);
+
+  *username = *password = *domain = NULL;
+  g_object_get (priv->session,
+                "username", username,
+                "password", password,
+                "domain", domain,
+                NULL);
+
+  if (*username != NULL && *username[0] == '\0' &&
+      *password != NULL && *password[0] == '\0' &&
+      *domain != NULL && *domain[0] == '\0')
+    return FALSE;
+
+  return TRUE;
+}
+
+/*
+ * frdp_display_authenticate_finish:
+ * @display: (transfer none): the RDP display widget
+ * @username: (transfer none): username for the connection
+ * @password: (transfer none): password for the connection
+ * @domain: (transfer none): optional domain for the connection
+ *
+ * This function finishes authentication which started in
+ * frdp_display_authenticate() and stores given authentication
+ * credentials into FrdpSession so that frdp_display_authenticate()
+ * can pick them up later and pass them to FreeRDP.
+ *
+ */
+void
+frdp_display_authenticate_finish (FrdpDisplay *self,
+                                  gchar       *username,
+                                  gchar       *password,
+                                  gchar       *domain)
+{
+  FrdpDisplayPrivate *priv = frdp_display_get_instance_private (self);
+
+  g_object_set (priv->session,
+                "username", username,
+                "password", password,
+                "domain", domain,
+                NULL);
+
+  priv->awaiting_authentication = FALSE;
+}
+
+guint
+frdp_display_certificate_verify_ex (FrdpDisplay *display,
+                                    const gchar *host,
+                                    guint16      port,
+                                    const gchar *common_name,
+                                    const gchar *subject,
+                                    const gchar *issuer,
+                                    const gchar *fingerprint,
+                                    DWORD        flags)
+{
+  FrdpDisplayPrivate *priv = frdp_display_get_instance_private (display);
+  GMainContext       *context;
+
+  g_signal_emit (display,
+                 signals[RDP_NEEDS_CERTIFICATE_VERIFICATION],
+                 0,
+                 host,
+                 port,
+                 common_name,
+                 subject,
+                 issuer,
+                 fingerprint,
+                 flags);
+
+  priv->awaiting_certificate_verification = TRUE;
+
+  context = g_main_context_default ();
+
+  while (priv->awaiting_certificate_verification)
+    g_main_context_iteration (context, FALSE);
+
+  return priv->certificate_verification_value;
+}
+
+guint
+frdp_display_certificate_change_verify_ex (FrdpDisplay *display,
+                                           const gchar *host,
+                                           guint16      port,
+                                           const gchar *common_name,
+                                           const gchar *subject,
+                                           const gchar *issuer,
+                                           const gchar *fingerprint,
+                                           const gchar *old_subject,
+                                           const gchar *old_issuer,
+                                           const gchar *old_fingerprint,
+                                           DWORD        flags)
+{
+  FrdpDisplayPrivate *priv = frdp_display_get_instance_private (display);
+  GMainContext       *context;
+
+  g_signal_emit (display, signals[RDP_NEEDS_CERTIFICATE_CHANGE_VERIFICATION], 0,
+                 host,
+                 port,
+                 common_name,
+                 subject,
+                 issuer,
+                 fingerprint,
+                 old_subject,
+                 old_issuer,
+                 old_fingerprint,
+                 flags);
+
+  priv->awaiting_certificate_change_verification = TRUE;
+
+  context = g_main_context_default ();
+
+  while (priv->awaiting_certificate_change_verification)
+    g_main_context_iteration (context, FALSE);
+
+  return priv->certificate_change_verification_value;
+}
+
+/**
+ * frdp_display_certificate_verify:
+ * @display: (transfer none): the RDP display widget
+ * @verification: verification value (1 - accept and store the certificate,
+ *                                    2 - accept the certificate for this session only
+ *                                    0 - otherwise)
+ *
+ * Finishes verification requested by FreeRDP.
+ */
+void
+frdp_display_certificate_verify_ex_finish (FrdpDisplay *display,
+                                           guint        verification)
+{
+  FrdpDisplayPrivate *priv = frdp_display_get_instance_private (display);
+
+  if (verification <= 2) {
+    priv->certificate_verification_value = verification;
+  }
+  else {
+    priv->certificate_verification_value = 0;
+    g_warning ("Verification value is out of allowed values.");
+  }
+  priv->awaiting_certificate_verification = FALSE;
+}
+
+/**
+ * frdp_display_certificate_change_verify:
+ * @display: (transfer none): the RDP display widget
+ * @verification: verification value (1 - accept and store the certificate,
+ *                                    2 - accept the certificate for this session only
+ *                                    0 - otherwise)
+ *
+ * Finishes verification requested by FreeRDP.
+ */
+void
+frdp_display_certificate_change_verify_ex_finish (FrdpDisplay *display,
+                                                  guint        verification)
+{
+  FrdpDisplayPrivate *priv = frdp_display_get_instance_private (display);
+
+  if (verification <= 2) {
+    priv->certificate_change_verification_value = verification;
+  }
+  else {
+    priv->certificate_change_verification_value = 0;
+    g_warning ("Verification value is out of allowed values.");
+  }
+  priv->awaiting_certificate_change_verification = FALSE;
 }
 
 /**
