@@ -51,6 +51,12 @@
 #define SELECT_TIMEOUT 50
 #define FRDP_CONNECTION_THREAD_MAX_ERRORS 10
 
+#ifdef HAVE_FREERDP3
+#define CONST_QUALIFIER const
+#else
+#define CONST_QUALIFIER
+#endif
+
 struct frdp_pointer
 {
 	rdpPointer pointer;
@@ -77,6 +83,7 @@ struct _FrdpSessionPrivate
   gchar *hostname;
   gchar *username;
   gchar *password;
+  gchar *domain;
   guint  port;
 
   gboolean show_cursor;
@@ -100,7 +107,8 @@ enum
   PROP_PASSWORD,
   PROP_DISPLAY,
   PROP_SCALING,
-  PROP_MONITOR_LAYOUT_SUPPORTED
+  PROP_MONITOR_LAYOUT_SUPPORTED,
+  PROP_DOMAIN
 };
 
 enum
@@ -257,10 +265,10 @@ frdp_session_configure_event (GtkWidget *widget,
   width = (double)gtk_widget_get_allocated_width (scrolled);
   height = (double)gtk_widget_get_allocated_height (scrolled);
 
-  if (priv->freerdp_session->settings == NULL)
+  if (priv->freerdp_session->context->settings == NULL)
     return;
 
-  settings = priv->freerdp_session->settings;
+  settings = priv->freerdp_session->context->settings;
 
   g_object_get (G_OBJECT (widget), "allow-resize", &allow_resize, NULL);
 
@@ -355,30 +363,58 @@ frdp_session_draw (GtkWidget *widget,
   return TRUE;
 }
 
+/*
+ * Return 1 to accept and store a certificate, 2 to accept
+ * a certificate only for this session, 0 otherwise.
+ */
 static guint
-frdp_certificate_verify (freerdp     *freerdp_session,
-                         const gchar *common_name,
-                         const gchar* subject,
-                         const gchar* issuer,
-                         const gchar* fingerprint,
-                         gboolean     host_mismatch)
+frdp_certificate_verify_ex (freerdp     *freerdp_session,
+                            const gchar *host,
+                            guint16      port,
+                            const gchar *common_name,
+                            const gchar *subject,
+                            const gchar *issuer,
+                            const gchar *fingerprint,
+                            guint32      flags)
 {
-  /* TODO */
-  return TRUE;
+  FrdpSession *self = ((frdpContext *) freerdp_session->context)->self;
+
+  return frdp_display_certificate_verify_ex (FRDP_DISPLAY (self->priv->display),
+                                             host,
+                                             port,
+                                             common_name,
+                                             subject,
+                                             issuer,
+                                             fingerprint,
+                                             flags);
 }
 
 static guint
-frdp_changed_certificate_verify (freerdp     *freerdp_session,
-                                 const gchar *common_name,
-                                 const gchar *subject,
-                                 const gchar *issuer,
-                                 const gchar *new_fingerprint,
-                                 const gchar *old_subject,
-                                 const gchar *old_issuer,
-                                 const gchar *old_fingerprint)
+frdp_changed_certificate_verify_ex (freerdp     *freerdp_session,
+                                    const gchar *host,
+                                    guint16      port,
+                                    const gchar *common_name,
+                                    const gchar *subject,
+                                    const gchar *issuer,
+                                    const gchar *fingerprint,
+                                    const gchar *old_subject,
+                                    const gchar *old_issuer,
+                                    const gchar *old_fingerprint,
+                                    guint32      flags)
 {
-  /* TODO */
-  return TRUE;
+  FrdpSession *self = ((frdpContext *) freerdp_session->context)->self;
+
+  return frdp_display_certificate_change_verify_ex (FRDP_DISPLAY (self->priv->display),
+                                                    host,
+                                                    port,
+                                                    common_name,
+                                                    subject,
+                                                    issuer,
+                                                    fingerprint,
+                                                    old_subject,
+                                                    old_issuer,
+                                                    old_fingerprint,
+                                                    flags);
 }
 
 static gboolean
@@ -405,8 +441,8 @@ caps_set (FrdpChannelDisplayControl *channel,
 }
 
 static void
-frdp_on_channel_connected_event_handler (void                      *context,
-                                         ChannelConnectedEventArgs *e)
+frdp_on_channel_connected_event_handler (void                                      *context,
+                                         CONST_QUALIFIER ChannelConnectedEventArgs *e)
 {
   frdpContext        *ctx = (frdpContext *) context;
   FrdpSession        *session = ctx->self;
@@ -447,8 +483,8 @@ frdp_on_channel_connected_event_handler (void                      *context,
 }
 
 static void
-frdp_on_channel_disconnected_event_handler (void                         *context,
-                                            ChannelDisconnectedEventArgs *e)
+frdp_on_channel_disconnected_event_handler (void                                         *context,
+                                            CONST_QUALIFIER ChannelDisconnectedEventArgs *e)
 {
   frdpContext        *ctx = (frdpContext *) context;
   FrdpSession        *session = ctx->self;
@@ -477,10 +513,19 @@ frdp_on_channel_disconnected_event_handler (void                         *contex
   }
 }
 
+static gboolean
+frdp_load_channels (freerdp *instance)
+{
+  if (!freerdp_client_load_addins (instance->context->channels, instance->context->settings))
+    return FALSE;
+
+  return TRUE;
+}
+
 static gboolean
 frdp_pre_connect (freerdp *freerdp_session)
 {
-  rdpSettings *settings = freerdp_session->settings;
+  rdpSettings *settings = freerdp_session->context->settings;
   rdpContext *context = freerdp_session->context;
 
   settings->OrderSupport[NEG_DSTBLT_INDEX] = TRUE;
@@ -513,10 +558,11 @@ frdp_pre_connect (freerdp *freerdp_session)
   PubSub_SubscribeChannelDisconnected (context->pubSub,
                                        frdp_on_channel_disconnected_event_handler);
 
-  if (!freerdp_client_load_addins (context->channels, settings))
-    return FALSE;
-
+#ifndef HAVE_FREERDP3
+  return frdp_load_channels (freerdp_session);
+#else
   return TRUE;
+#endif
 }
 
 static gboolean
@@ -577,7 +623,7 @@ frdp_post_connect (freerdp *freerdp_session)
   settings = context->settings;
   switch (frdp_session_get_best_color_depth (self)) {
     case 32:
-      color_format = PIXEL_FORMAT_BGRA32;
+      color_format = PIXEL_FORMAT_BGRX32;
       self->priv->cairo_format = CAIRO_FORMAT_ARGB32;
       break;
     case 24:
@@ -597,9 +643,9 @@ frdp_post_connect (freerdp *freerdp_session)
 
   gdi_init (freerdp_session, color_format);
 
-  freerdp_session->update->BeginPaint = frdp_begin_paint;
-  freerdp_session->update->EndPaint = frdp_end_paint;
-  freerdp_session->update->DesktopResize = frdp_desktop_resize;
+  freerdp_session->context->update->BeginPaint = frdp_begin_paint;
+  freerdp_session->context->update->EndPaint = frdp_end_paint;
+  freerdp_session->context->update->DesktopResize = frdp_desktop_resize;
 
   EventArgsInit(&e, "frdp");
 	e.width = settings->DesktopWidth;
@@ -691,14 +737,215 @@ update (gpointer user_data)
   return TRUE;
 }
 
+/*
+ *  The local layout string is taken from org.gnome.desktop.input-sources.mru-sources
+ *  after adding it among available sources and setting it active.
+ *  Most of the national layout has been assigned according to default keyboard
+ *  for given nation so these can be inaccurate.
+ *  Non-standard layouts has to be checked directly by comparing them
+ *  (e.g. against layouts on https://kbdlayout.info/).
+ */
+static const struct {
+  const char  *local_layout;
+  const guint  freerdp_layout;
+} keyboard_layouts[] = {
+  { "ara", KBD_ARABIC_101 },
+  { "bg", KBD_BULGARIAN },
+  { "cz", KBD_CZECH },
+  { "da", KBD_DANISH },
+  { "de", KBD_GERMAN },
+  { "gr", KBD_GREEK },
+  { "us", KBD_US },
+  { "es", KBD_SPANISH },
+  { "fi", KBD_FINNISH },
+  { "fr", KBD_FRENCH },
+  { "il", KBD_HEBREW },
+  { "hu", KBD_HUNGARIAN },
+  { "is", KBD_ICELANDIC },
+  { "it", KBD_ITALIAN },
+  { "jp", KBD_JAPANESE },
+  { "kr", KBD_KOREAN },
+  { "nl", KBD_DUTCH },
+  { "no", KBD_NORWEGIAN },
+  { "pl+dvp", KBD_POLISH_PROGRAMMERS },
+  { "pt", KBD_PORTUGUESE_BRAZILIAN_ABNT },
+  { "ro", KBD_ROMANIAN },
+  { "ru", KBD_RUSSIAN },
+  { "hr", KBD_CROATIAN },
+  { "sk", KBD_SLOVAK },
+  { "al", KBD_ALBANIAN },
+  { "se", KBD_SWEDISH },
+  { "tr", KBD_TURKISH_Q },
+  { "pk+urd-nla", KBD_URDU },
+  { "ua", KBD_UKRAINIAN },
+  { "by", KBD_BELARUSIAN },
+  { "si", KBD_SLOVENIAN },
+  { "ee", KBD_ESTONIAN },
+  { "lv", KBD_LATVIAN },
+  { "lt+ibm", KBD_LITHUANIAN_IBM },
+  { "vn", KBD_VIETNAMESE },
+  { "am+eastern", KBD_ARMENIAN_EASTERN },
+  { "mk", KBD_FYRO_MACEDONIAN },
+  { "ge", KBD_GEORGIAN },
+  { "fo", KBD_FAEROESE },
+  { "no+smi", KBD_NORWEGIAN_WITH_SAMI },
+  { "kz", KBD_KAZAKH },
+  { "gb+extd", KBD_UNITED_KINGDOM_EXTENDED },
+  { "sy+syc", KBD_SYRIAC },
+  { "af+ps-olpc", KBD_PASHTO },
+  { "m17n:dv:phonetic", KBD_DIVEHI_PHONETIC },
+  { "nz+mao", KBD_MAORI },
+  { "ch", KBD_SWISS_GERMAN },
+  { "gb", KBD_UNITED_KINGDOM },
+  { "latam", KBD_LATIN_AMERICAN },
+  { "be", KBD_BELGIAN_FRENCH },
+  { "pt", KBD_PORTUGUESE },
+  { "ca+fr-legacy", KBD_CANADIAN_FRENCH_LEGACY },
+  { "ca", KBD_CANADIAN_FRENCH },
+  { "ba", KBD_BOSNIAN },
+  { "cz+qwerty", KBD_CZECH_QWERTY },
+  { "de+T3", KBD_GERMAN_IBM },
+  { "us+dvorak", KBD_UNITED_STATES_DVORAK },
+  { "it+ibm", KBD_ITALIAN_142 },
+  { "pl+qwertz", KBD_POLISH_214 },
+  { "pt", KBD_PORTUGUESE_BRAZILIAN_ABNT2 },
+  { "sk+qwerty", KBD_SLOVAK_QWERTY },
+  { "th+pat", KBD_THAI_PATTACHOTE },
+  { "tr+f", KBD_TURKISH_F },
+  { "lt", KBD_LITHUANIAN },
+  { "am+western", KBD_ARMENIAN_WESTERN },
+  { "be+iso-alternate", KBD_BELGIAN_COMMA },
+  { "gb+gla", KBD_GAELIC },
+  { "us+intl", KBD_UNITED_STATES_INTERNATIONAL },
+  { "us+dvorak-l", KBD_UNITED_STATES_DVORAK_FOR_LEFT_HAND },
+  { "us+dvorak-r", KBD_UNITED_STATES_DVORAK_FOR_RIGHT_HAND },
+  { "us+dvp", KBD_UNITED_STATES_DVORAK_PROGRAMMER },
+  { "gr+polytonic", KBD_GREEK_POLYTONIC },
+  { "fr+bepo", KBD_FRENCH_BEPO },
+  { "de+neo", KBD_GERMAN_NEO },
+
+  /* These need to be determined yet. */
+
+  { "", KBD_CHINESE_TRADITIONAL_US },
+  { "", KBD_THAI_KEDMANEE },
+  { "", KBD_FARSI },
+  { "", KBD_AZERI_LATIN },
+  { "", KBD_DEVANAGARI_INSCRIPT },
+  { "", KBD_MALTESE_47_KEY },
+  { "", KBD_KYRGYZ_CYRILLIC },
+  { "", KBD_TATAR },
+  { "", KBD_BENGALI },
+  { "", KBD_PUNJABI },
+  { "", KBD_GUJARATI },
+  { "", KBD_TAMIL },
+  { "", KBD_TELUGU },
+  { "", KBD_KANNADA },
+  { "", KBD_MALAYALAM },
+  { "", KBD_MARATHI },
+  { "", KBD_MONGOLIAN_CYRILLIC },
+  { "", KBD_NEPALI },
+  { "", KBD_LUXEMBOURGISH },
+  { "", KBD_CHINESE_SIMPLIFIED_US },
+  { "", KBD_BELGIAN_PERIOD },
+  { "", KBD_SERBIAN_LATIN },
+  { "", KBD_AZERI_CYRILLIC },
+  { "", KBD_SWEDISH_WITH_SAMI },
+  { "", KBD_UZBEK_CYRILLIC },
+  { "", KBD_INUKTITUT_LATIN },
+  { "", KBD_SERBIAN_CYRILLIC },
+  { "", KBD_SWISS_FRENCH },
+  { "", KBD_IRISH },
+  { "", KBD_BOSNIAN_CYRILLIC },
+  { "", KBD_ARABIC_102 },
+  { "", KBD_BULGARIAN_LATIN },
+  { "", KBD_GREEK_220 },
+  { "", KBD_SPANISH_VARIATION },
+  { "", KBD_HUNGARIAN_101_KEY },
+  { "", KBD_RUSSIAN_TYPEWRITER },
+  { "", KBD_LATVIAN_QWERTY },
+  { "", KBD_HINDI_TRADITIONAL },
+  { "", KBD_MALTESE_48_KEY },
+  { "", KBD_SAMI_EXTENDED_NORWAY },
+  { "", KBD_BENGALI_INSCRIPT },
+  { "", KBD_SYRIAC_PHONETIC },
+  { "", KBD_DIVEHI_TYPEWRITER },
+  { "", KBD_FINNISH_WITH_SAMI },
+  { "", KBD_CANADIAN_MULTILINGUAL_STANDARD },
+  { "", KBD_ARABIC_102_AZERTY },
+  { "", KBD_CZECH_PROGRAMMERS },
+  { "", KBD_GREEK_319 },
+  { "", KBD_THAI_KEDMANEE_NON_SHIFTLOCK },
+  { "", KBD_SAMI_EXTENDED_FINLAND_SWEDEN },
+  { "", KBD_GREEK_220_LATIN },
+  { "", KBD_THAI_PATTACHOTE_NON_SHIFTLOCK },
+  { "", KBD_GREEK_319_LATIN },
+  { "", KBD_GREEK_LATIN },
+  { "", KBD_US_ENGLISH_TABLE_FOR_IBM_ARABIC_238_L }
+
+/*
+  These are available in FreeRDP 3:
+  { "", KBD_ROMANIAN_STANDARD },
+  { "", KBD_GEORGIAN_QUERTY },
+  { "", KBD_KHMER },
+  { "", KBD_RUSSIAN_PHONETIC },
+  { "", KBD_BANGLA },
+  { "", KBD_BULGARIAN_PHONETIC },
+  { "", KBD_PERSIAN },
+*/
+};
+
+static void
+frdp_session_set_current_keyboard_layout (FrdpSession *self) {
+  GSettingsSchemaSource *source;
+  FrdpSessionPrivate    *priv = self->priv;
+  GSettingsSchema       *schema;
+  rdpSettings           *settings;
+  const gchar           *source_type = NULL, *layout = NULL;
+  GSettings             *gsettings;
+  GVariant              *sources;
+  gboolean               keyboard_layout_set = FALSE;
+  guint                  i;
+
+  settings = priv->freerdp_session->context->settings;
+
+  source = g_settings_schema_source_get_default ();
+  if (source != NULL) {
+    schema = g_settings_schema_source_lookup (source, "org.gnome.desktop.input-sources", TRUE);
+    if (schema != NULL) {
+      gsettings = g_settings_new (g_settings_schema_get_id (schema));
+      sources = g_settings_get_value (gsettings, "mru-sources");
+
+      if (g_variant_n_children (sources) >= 1)
+        g_variant_get_child (sources, 0, "(&s&s)", &source_type, &layout);
+
+      if (layout != NULL) {
+        for (i = 0; i < G_N_ELEMENTS (keyboard_layouts); i++) {
+          if (g_strcmp0 (layout, keyboard_layouts[i].local_layout) == 0) {
+            settings->KeyboardLayout = freerdp_keyboard_init (keyboard_layouts[i].freerdp_layout);
+            keyboard_layout_set = TRUE;
+            break;
+          }
+        }
+      }
+
+      g_variant_unref (sources);
+      g_object_unref (gsettings);
+      g_settings_schema_unref (schema);
+    }
+  }
+
+  if (!keyboard_layout_set)
+    settings->KeyboardLayout = freerdp_keyboard_init (0);
+}
+
 static void
 frdp_session_init_freerdp (FrdpSession *self)
 {
-  FrdpSessionPrivate *priv = self->priv;
-  rdpSettings        *settings;
-  gchar              *collections[1];
-  gchar              *build_options;
-  int                 count = 1;
+  CONST_QUALIFIER gchar *collections[] = { "disp" };
+  FrdpSessionPrivate    *priv = self->priv;
+  rdpSettings           *settings;
+  gchar                 *build_options;
+  int                    count = 1;
 
   /* Setup FreeRDP session */
   priv->freerdp_session = freerdp_new ();
@@ -706,20 +953,24 @@ frdp_session_init_freerdp (FrdpSession *self)
   priv->freerdp_session->PostConnect = frdp_post_connect;
   priv->freerdp_session->PostDisconnect = frdp_post_disconnect;
   priv->freerdp_session->Authenticate = frdp_authenticate;
-  priv->freerdp_session->VerifyCertificate = frdp_certificate_verify;
-  priv->freerdp_session->VerifyChangedCertificate = frdp_changed_certificate_verify;
+  priv->freerdp_session->VerifyCertificateEx = frdp_certificate_verify_ex;
+  priv->freerdp_session->VerifyChangedCertificateEx = frdp_changed_certificate_verify_ex;
+#ifdef HAVE_FREERDP3
+  priv->freerdp_session->LoadChannels = frdp_load_channels;
+#endif
 
   priv->freerdp_session->ContextSize = sizeof (frdpContext);
 
   freerdp_context_new (priv->freerdp_session);
   ((frdpContext *) priv->freerdp_session->context)->self = self;
 
-  settings = priv->freerdp_session->settings;
+  settings = priv->freerdp_session->context->settings;
 
   settings->ServerHostname = g_strdup (priv->hostname);
   settings->ServerPort = priv->port;
   settings->Username = g_strdup (priv->username);
   settings->Password = g_strdup (priv->password);
+  settings->Domain = g_strdup (priv->domain);
 
   settings->AllowFontSmoothing = TRUE;
   settings->AllowUnanouncedOrdersFromServer = TRUE;
@@ -742,7 +993,6 @@ frdp_session_init_freerdp (FrdpSession *self)
   settings->RedirectClipboard = TRUE;
   settings->SupportGraphicsPipeline = TRUE;
 
-  collections[0] = "disp";
   freerdp_client_add_dynamic_channel (settings, count, collections);
 
   build_options = g_ascii_strup (freerdp_get_build_config (), -1);
@@ -755,7 +1005,7 @@ frdp_session_init_freerdp (FrdpSession *self)
   }
   g_free (build_options);
 
-  settings->KeyboardLayout = freerdp_keyboard_init (0);
+  frdp_session_set_current_keyboard_layout (self);
 
   freerdp_register_addin_provider(freerdp_channels_load_static_addin_entry, 0);
 }
@@ -833,7 +1083,7 @@ frdp_session_get_property (GObject    *object,
                            GParamSpec *pspec)
 {
   FrdpSession *self = (FrdpSession*) object;
-  rdpSettings *settings = self->priv->freerdp_session->settings;
+  rdpSettings *settings = self->priv->freerdp_session->context->settings;
 
   switch (property_id)
     {
@@ -849,6 +1099,9 @@ frdp_session_get_property (GObject    *object,
       case PROP_PASSWORD:
         g_value_set_string (value, settings->Password);
         break;
+      case PROP_DOMAIN:
+        g_value_set_string (value, settings->Domain);
+        break;
       case PROP_DISPLAY:
         g_value_set_object (value, self->priv->display);
         break;
@@ -889,6 +1142,10 @@ frdp_session_set_property (GObject      *object,
         g_free (self->priv->password);
         self->priv->password = g_value_dup_string (value);
         break;
+      case PROP_DOMAIN:
+        g_free (self->priv->domain);
+        self->priv->domain = g_value_dup_string (value);
+        break;
       case PROP_DISPLAY:
         self->priv->display = g_value_get_object (value);
         break;
@@ -956,6 +1213,14 @@ frdp_session_class_init (FrdpSessionClass *klass)
                                                         NULL,
                                                         G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE));
 
+  g_object_class_install_property (gobject_class,
+                                   PROP_DOMAIN,
+                                   g_param_spec_string ("domain",
+                                                        "domain",
+                                                        "domain",
+                                                        NULL,
+                                                        G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE));
+
   g_object_class_install_property (gobject_class,
                                    PROP_DISPLAY,
                                    g_param_spec_object ("display",
@@ -1111,7 +1376,7 @@ frdp_session_mouse_event (FrdpSession          *self,
   if (event & FRDP_MOUSE_EVENT_BUTTON5)
     xflags |=  PTR_XFLAGS_BUTTON2;
 
-  input = priv->freerdp_session->input;
+  input = priv->freerdp_session->context->input;
 
   if (priv->scaling) {
     x = (x - priv->offset_x) / priv->scale;
@@ -1129,6 +1394,57 @@ frdp_session_mouse_event (FrdpSession          *self,
   }
 }
 
+void
+frdp_session_mouse_smooth_scroll_event (FrdpSession          *self,
+                                        guint16               x,
+                                        guint16               y,
+                                        double                delta_x,
+                                        double                delta_y)
+{
+  FrdpSessionPrivate *priv = self->priv;
+  rdpInput           *input;
+  guint16             flags = 0, value;
+
+  g_return_if_fail (priv->freerdp_session != NULL);
+
+  if (fabs (delta_y) >= fabs (delta_x)) {
+    flags |= PTR_FLAGS_WHEEL;
+    value = (guint16) round (fabs (delta_y) * 0x78);
+    if (value > 0) {
+      /* Reversing direction here to reflect the behaviour on local side. */
+      if (delta_y < 0.0) {
+        if (value > 255)
+          value = 255;
+        flags |= value & WheelRotationMask;
+      } else {
+        if (value > 256)
+          value = 256;
+        flags |= PTR_FLAGS_WHEEL_NEGATIVE;
+        flags |= (~value + 1) & WheelRotationMask;
+      }
+    }
+  } else {
+    flags |= PTR_FLAGS_HWHEEL;
+    value = (guint16) round (fabs (delta_x) * 0x78);
+    if (value > 0) {
+      if (delta_x < 0.0) {
+        if (value > 256)
+          value = 256;
+        flags |= PTR_FLAGS_WHEEL_NEGATIVE;
+        flags |= (~value + 1) & WheelRotationMask;
+      } else {
+        if (value > 255)
+          value = 255;
+        flags |= value & WheelRotationMask;
+      }
+    }
+  }
+
+  input = priv->freerdp_session->context->input;
+
+  freerdp_input_send_mouse_event (input, flags, 0, 0);
+}
+
 void
 frdp_session_mouse_pointer  (FrdpSession          *self,
                              gboolean              enter)
@@ -1143,7 +1459,7 @@ void
 frdp_session_send_key (FrdpSession  *self,
                        GdkEventKey  *key)
 {
-  rdpInput *input = self->priv->freerdp_session->input;
+  rdpInput *input = self->priv->freerdp_session->context->input;
   DWORD scancode = 0;
   guint8 keycode;
   guint16 flags;
